/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Message;
import javax.wsdl.Port;
import javax.wsdl.PortType;
import javax.wsdl.Service;
import javax.xml.namespace.QName;

import org.jbpm.pvm.env.Environment;
import org.jbpm.pvm.internal.job.TimerImpl;
import org.jbpm.pvm.internal.model.ExecutionImpl;
import org.jbpm.pvm.model.OpenExecution;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.facade.data.ProcessInstanceData;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.Receiver;
import org.ow2.orchestra.services.ReplierKey;
import org.ow2.orchestra.services.itf.ReceiverRepository;
import org.ow2.orchestra.services.itf.Replier;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.wsdl.PartnerLinkType;
import org.ow2.orchestra.wsdl.Property;
import org.ow2.orchestra.wsdl.PropertyAlias;
import org.ow2.orchestra.wsdl.WsdlsInfos;

/**
 * a runtime path of execution.
 *
 * @author Tom Baeyens, Charles Souillard
 */
public class BpelExecution extends ExecutionImpl {

  public static final String ADDITIONNAL_SCOPE_VARIABLES = "additionnalScopeVariables";

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  protected ProcessInstanceUUID processInstanceUUID;

  protected ActivityInstanceUUID activityInstanceUUID;

  /**
   * This variable is meaningful only in bpel execution which have no parent
   */
  protected long activityCount;

  /** log. */
  private static Logger log = Logger.getLogger(BpelExecution.class.getName());

  protected List<FlowRuntime> flowRuntimes = new ArrayList<FlowRuntime>();

  protected ForEachRuntime forEachRuntime;

  protected List<ScopeRuntime> scopeRuntimes = new ArrayList<ScopeRuntime>();

  protected String uuid;

  protected boolean hasPendingMessage;

  protected Set<BpelExecution> waitingExecutions;

  /**
   * This variable has same value in one hierarchy.
   */
  protected long instancePosition;

  /**
   * This variable is meaningful in each bpel execution (activityCount-1 is not always the value because
   * we can have parallel activities).
   */
  protected long activityPosition;

  protected BpelExecution() {
    super();
  }

  public BpelExecution(final ProcessInstanceUUID processInstanceUUID, final String uuid, final long instancePosition) {
    this.instancePosition = instancePosition;
    this.uuid = uuid;
    this.processInstanceUUID = processInstanceUUID;
    this.activityCount = 0;
  }

  @Override
  protected BpelExecution newChildExecution() {
    final BpelExecution child = new BpelExecution(this.processInstanceUUID, UUID.randomUUID().toString(), this.instancePosition);
    child.setActivityInstanceUUID(this.activityInstanceUUID);
    child.setActivityPosition(this.activityPosition);
    return child;
  }

  @Override
  public BpelExecution getParent() {
    return (BpelExecution) super.getParent();
  }

  /**
   * Returns the rootExecution (instance).
   *
   * @return the rootExecution (instance)
   */
  @Override
  public BpelExecution getProcessInstance() {
    return (BpelExecution) super.getProcessInstance();
  }

  /**
   * Search for the flowRuntime having a link declared with the given name. We
   * start searching into parent and not into this execution as links used in an
   * activity (the current node) must be declared before (in an enclosing flow).
   * So if the current node is flow, we don't want to search into its links. If
   * this parent execution flowRuntime doesn't have one, look into parent
   *
   * @param linkName
   *            linkName
   * @return the flowRuntime having a link declared with the given name.
   */
  public FlowRuntime getFlowRuntimeWithLink(final String linkName) {
    final List<FlowRuntime> fRuntimes = getFlowRuntimes();
    if (fRuntimes != null) {
      for (final FlowRuntime fr : fRuntimes) {
        if (fr.hasLink(linkName)) {
          return fr;
        }
      }
    }
    if (this.parent != null) {
      return (getParent()).getFlowRuntimeWithLink(linkName);
    }
    return null;
  }

  public ScopeRuntime getScopeRuntime() {
    if (this.scopeRuntimes != null && !this.scopeRuntimes.isEmpty()) {
      return this.scopeRuntimes.get(0);
    }
    if (this.parent != null) {
      return (getParent()).getScopeRuntime();
    }
    return null;
  }

  public ScopeRuntime getScopeRuntimeWithVariable(final String variableName) {
    if (this.scopeRuntimes != null) {
      for (final ScopeRuntime sr : this.scopeRuntimes) {
        if (sr.containsVariable(variableName)) {
          return sr;
        }
      }
    }
    if (this.parent != null) {
      return (getParent()).getScopeRuntimeWithVariable(variableName);
    }
    return null;
  }

  public ScopeRuntime getScopeRuntimeWithCS(final String csName) {
    if (this.scopeRuntimes != null) {
      for (final ScopeRuntime sr : this.scopeRuntimes) {
        if (sr.containsCorrelationSet(csName)) {
          return sr;
        }
      }
    }
    if (this.parent != null) {
      return (getParent()).getScopeRuntimeWithCS(csName);
    }
    return null;
  }

  public ScopeRuntime getScopeRuntimeWithPLNK(final String partnerLinkName) {
    if (this.scopeRuntimes != null) {
      for (final ScopeRuntime sr : this.scopeRuntimes) {
        if (sr.getPartnerLink(partnerLinkName) != null) {
          return sr;
        }
      }
    }
    if (this.parent != null) {
      return (getParent()).getScopeRuntimeWithPLNK(partnerLinkName);
    }
    return null;
  }

  /**
   * Set the given linkStatus to the first enclosing flow having a link with the
   * given linkName.
   *
   * @param linkName
   *            linkName of the searched link
   * @param linkStatus
   *            linkStatus to set
   */
  public void setLinkStatus(final String linkName, final Boolean linkStatus) {
    final FlowRuntime fr = getFlowRuntimeWithLink(linkName);
    fr.setLinkStatus(linkName, linkStatus);
  }

  /**
   * toString method.
   *
   * @return String string representation of this execution.
   */
  @Override
  public String toString() {
    return "BpelExecution" + this.hashCode();
  }

  /**
   * The process of this execution is guaranteed to be a {@link BpelProcess}.
   */
  @Override
  public BpelProcess getProcessDefinition() {
    return (BpelProcess) this.processDefinition;
  }

  /**
   * Search in currentScope (and parents) for a partnerLinkRuntime having the
   * given name.
   *
   * @param partnerLinkName
   *            partnerLinkName of the searched partnerLinkRuntime
   * @return the partnerLinkRuntime having the given partnerLinkName in the
   *         closest enclosing scope.
   */
  public PartnerLinkRuntime getPartnerLinkRuntime(final String partnerLinkName) {
    final ScopeRuntime scopeRuntime = getScopeRuntimeWithPLNK(partnerLinkName);
    if (scopeRuntime != null) {
      return scopeRuntime.getPartnerLink(partnerLinkName);
    }
    return null;
  }

  /**
   *
   * /** Search for a variableRuntime having the given variableName in the
   * closest scope and return its value.
   *
   * @param variableName
   *            variableName
   * @return Search for a variableRuntime having the given variableName in the
   *         closest scope and return its value.
   */
  public Object getVariableValue(final String variableName) {
    final VariableRuntime variable = getVariableByName(variableName);
    return variable.getValue();
  }

  /**
   * Search for a variableRuntime having the given variableName in the closest
   * scope.
   *
   * @param variableName
   *            variableName
   * @return the found variableRuntime
   */
  public VariableRuntime getVariableByName(final String variableName) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("Getting variable " + variableName);
    }
    final ScopeRuntime currentRuntimeScope = getScopeRuntimeWithVariable(variableName);
    if (currentRuntimeScope != null) {
      return currentRuntimeScope.getVariable(variableName, this);
    }
    return null;
  }

  /**
   * Search for a correlationSetRuntime having the given csName in the closest
   * scope.
   *
   * @param csName
   *            csName
   * @return the found correlationSetRuntime
   */
  public CorrelationSetRuntime getCorrelationSetByName(final String csName) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("Getting CorrelationSet " + csName);
    }
    final ScopeRuntime scopeRuntime = getScopeRuntimeWithCS(csName);
    if (scopeRuntime != null) {
      return scopeRuntime.getCorrelationSet(csName);
    }
    return null;
  }

  public void addPartnerLinkType(final PartnerLinkType partnerLinkType) {
    getWsdlInfos().addPartnerLinkType(partnerLinkType);
  }

  public PartnerLinkType getPartnerLinkType(final QName partnerLinkTypeName) {
    return getWsdlInfos().getPartnerLinkType(partnerLinkTypeName);
  }

  public void addPortTypes(final Map<QName, PortType> portTypes) {
    getWsdlInfos().addPortTypes(portTypes);
  }

  public PortType getPortType(final QName portTypeQName) {
    return getWsdlInfos().getPortType(portTypeQName);
  }

  public void addMessages(final Map<QName, Message> messages) {
    getWsdlInfos().addMessages(messages);
  }

  public Message getMessage(final QName messageName) {
    return getWsdlInfos().getMessage(messageName);
  }

  public void addServices(final Map<QName, Service> services) {
    getWsdlInfos().addServices(services);
  }

  public Service getService(final QName serviceName) {
    return getWsdlInfos().getService(serviceName);
  }

  public Service getServiceOfPort(final Port port) {
    return getWsdlInfos().getServiceOfPort(port);
  }

  public List<Port> getPorts(final QName portTypeQName) {
    return getWsdlInfos().getPorts(portTypeQName);
  }

  public void addProperty(final Property property) {
    getWsdlInfos().addProperty(property);
  }

  public Property getProperty(final QName propertyQName) {
    return getWsdlInfos().getProperty(propertyQName);
  }

  public Map<String, URL> getStyleSheets() {
    return getProcessDefinition().getStyleSheetRepository()
    .getStyleSheets();
  }

  public void addPropertyAlias(final PropertyAlias propertyAlias) {
    getWsdlInfos().addPropertyAlias(propertyAlias);
  }

  public PropertyAlias getPropertyAliasWithMessageType(final QName propertyName,
      final QName messageType) {
    return getWsdlInfos().getPropertyAliasWithMessageType(propertyName,
        messageType);
  }

  public List<PropertyAlias> getPropertyAliasesWithMessageType(final QName messageType) {
    return getWsdlInfos().getPropertyAliasesWithMessageType(messageType);
  }

  public List<PropertyAlias> getPropertyAliases(final CorrelationSet cs,
      final QName messageType) {
    return getWsdlInfos().getPropertyAliases(cs.getProperties().values(),
        messageType);
  }

  public PropertyAlias getPropertyAliasWithElement(final QName propertyName,
      final QName element) {
    return getWsdlInfos().getPropertyAliasWithElement(propertyName, element);
  }

  public PropertyAlias getPropertyAliasWithType(final QName propertyName, final QName type) {
    return getWsdlInfos().getPropertyAliasWithType(propertyName, type);
  }

  public WsdlsInfos getWsdlInfos() {
    return getProcessDefinition().getWsdlInfos();
  }

  public void setWsdlInfos(final WsdlsInfos wsdlInfos) {
    getProcessDefinition().setWsdlInfos(wsdlInfos);
  }



  public void pushFlowRuntime(final FlowRuntime flowRuntime) {
    if (this.flowRuntimes == null) {
      this.flowRuntimes = new ArrayList<FlowRuntime>();
    }
    this.flowRuntimes.add(0, flowRuntime);
  }

  public void popFlowRuntime() {
    if (this.flowRuntimes != null && !this.flowRuntimes.isEmpty()) {
      this.flowRuntimes.remove(0);
    }
  }

  public List<FlowRuntime> getFlowRuntimes() {
    return this.flowRuntimes;
  }

  /**
   * @param forEachRuntime the forEachRuntime to push
   */
  public void pushForEachRuntime(final ForEachRuntime forEachRuntime) {
    if (this.forEachRuntime != null) {
      throw new Error("More than one foreach runtime in one execution");
    }
    this.forEachRuntime = forEachRuntime;
  }

  /**
   * Pop a forEachRuntime.
   */
  public void popForEachRuntime() {
    this.forEachRuntime = null;
  }

  /**
   * Pop a forEachRuntime.
   */
  public ForEachRuntime getForEachRuntime() {
    return this.forEachRuntime;
  }


  public void pushScopeRuntime(final ScopeRuntime scopeRuntime) {
    if (this.scopeRuntimes == null) {
      this.scopeRuntimes = new ArrayList<ScopeRuntime>();
    }
    this.scopeRuntimes.add(0, scopeRuntime);
  }

  public ScopeRuntime popScopeRuntime() {
    ScopeRuntime res = null;
    if (this.scopeRuntimes != null && !this.scopeRuntimes.isEmpty()) {
      res = this.scopeRuntimes.remove(0);
    }
    return res;
  }

  public List<ScopeRuntime> getScopeRuntimes() {
    return this.scopeRuntimes;
  }

  public Replier getReplier(final OperationKey operationKey, final String messageExchange) {
    Replier replier = null;
    if (this.scopeRuntimes != null) {
      for (final ScopeRuntime sr : this.scopeRuntimes) {
        final ReplierKey replierKey = new ReplierKey(operationKey, messageExchange, sr);
        replier = Environment.getCurrent().get(Repository.class).removeReplier(
            replierKey);
        if (replier != null) {
          return replier;
        }
      }
    }
    if (this.parent != null) {
      return (getParent()).getReplier(operationKey, messageExchange);
    }
    return null;
  }

  public String getUUID() {
    return this.uuid;
  }

  @Override
  public int hashCode() {
    return this.uuid.hashCode();
  }


  @Override
  public boolean equals(final Object obj) {
    if (obj == null) {
      return false;
    }
    if (!(obj instanceof BpelExecution)) {
      return obj.equals(this);
    }
    final BpelExecution other = (BpelExecution) obj;
    return other.getUUID().equals(this.uuid);
  }

  public boolean hasPendingMessage() {
    return this.hasPendingMessage;
  }

  public void setHasPendingMessage(final boolean hasPendingMessage) {
    this.hasPendingMessage = hasPendingMessage;
  }

  public Set<BpelExecution> getWaitingExecutions() {
    return this.waitingExecutions;
  }

  public void addWaitingExecution(final BpelExecution execution) {
    if (this.waitingExecutions == null) {
      this.waitingExecutions = new HashSet<BpelExecution>();
    }
    this.waitingExecutions.add(execution);
  }

  @Override
  // TODO: if timers can be created by using TimerDefinition objects, this method should be removed.
  public void createTimer(final String eventName,
      final String signalName,
      final String dueDateDescription,
      final Date dueDate,
      final String repeat,
      final Boolean isExclusive,
      final Integer retries) {
    if (this.timers == null) {
      this.timers = new HashSet<TimerImpl>();
    }
    super.createTimer(eventName, signalName, dueDateDescription, dueDate, repeat, isExclusive, retries);
  }

  @Override
  protected TimerImpl instantiateTimer() {
    return new BpelTimerImpl();
  }

  public void terminate() {
    // Terminate children
    final Collection<OpenExecution> childs = this.getExecutions();
    if (childs != null) {
      for (final OpenExecution child : new ArrayList<OpenExecution>(childs)) {
        final BpelExecution bpelChild = (BpelExecution) child;
        bpelChild.terminate();
      }
    }
    // delete execution form all receivers
    final ReceiverRepository receiverRepository = Environment.getCurrent().get(ReceiverRepository.class);
    final QName processQName = getProcessDefinition().getQName();
    final List<Receiver> receivers = receiverRepository.getReceivers(processQName);
    if (receivers != null) {
      for (final Receiver receiver : receivers) {
        receiver.removeWaitingExecution(this);
      }
    }
    getProcessInstance().getWaitingExecutions().remove(this);
    if (this.parent != null) {
      this.parent.removeExecution(this);
    }
  }

  public ProcessInstanceUUID getProcessInstanceUUID() {
    return this.processInstanceUUID;
  }

  /**
   * @return the activityInstanceUUID
   */
  public ActivityInstanceUUID getActivityInstanceUUID() {
    return this.activityInstanceUUID;
  }

  /**
   * @return the activityInstanceUUID
   */
  public ActivityInstanceUUID getEncloserActivityInstanceUUID() {
    return getEncloserActivityInstanceUUID(getActivityInstanceUUID());
  }

  /**
   * @return the activityInstanceUUID
   */
  private ActivityInstanceUUID getEncloserActivityInstanceUUID(final ActivityInstanceUUID enclosedActivityUUID) {
    final ActivityInstanceUUID activityInstanceUUID = getActivityInstanceUUID();
    final BpelExecution parentBpelExecution = getParent();
    if (enclosedActivityUUID.equals(activityInstanceUUID)) {
      if (parentBpelExecution == null) {
        return null;
      } else {
        return parentBpelExecution.getEncloserActivityInstanceUUID(enclosedActivityUUID);
      }
    } else {
      return activityInstanceUUID;
    }
  }

  /**
   * @param activityInstanceUUID the activityInstanceUUID to set
   */
  public void setActivityInstanceUUID(final ActivityInstanceUUID activityInstanceUUID) {
    this.activityInstanceUUID = activityInstanceUUID;
  }

  public long getNextActivityCount() {
    if (getParent() == null) {
      this.activityCount++;
      return this.activityCount;
    } else {
      return getProcessInstance().getNextActivityCount();
    }
  }

  public ProcessInstanceData getProcessInstanceData() {
    return new ProcessInstanceData(getProcessDefinition().getProcessDefinitionData(), getProcessInstanceUUID());
  }

  /**
   * @return the activityPosition
   */
  public long getActivityPosition() {
    return this.activityPosition;
  }

  /**
   * @param activityPosition the activityPosition to set
   */
  public void setActivityPosition(final long activityPosition) {
    this.activityPosition = activityPosition;
  }

}
