/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.locks.ReentrantLock;

import javax.xml.namespace.QName;

import org.ow2.orchestra.runtime.BpelExecution;

public final class MutexRepository {

  public static class ProcessLock {
    private ReentrantLock lock = new ReentrantLock();

    private void lock() {
      lock.lock();
    }

    private void unlock() {
      lock.unlock();
    }
  }

  public static class InstanceLock {
    private ReentrantLock lock = new ReentrantLock();

    private void lock() {
      lock.lock();
    }

    private void unlock() {
      lock.unlock();
    }
  }

  private static Map<QName, ProcessLock> processMutexs = new HashMap<QName, ProcessLock>();

  private static Map<String, InstanceLock> instanceMutexs = new HashMap<String, InstanceLock>();

  private MutexRepository() {
  }


  private static synchronized ProcessLock getProcessLock(final QName processQName) {
    return processMutexs.get(processQName);
  }

  public static ProcessLock lockProcess(final QName processQName) {
    ProcessLock processLock = getProcessLock(processQName);
    if (processLock != null) {
      processLock.lock();
    }
    return processLock;
  }

  public static void unlockProcess(final ProcessLock processLock) {
    if (processLock != null) {
      processLock.unlock();
    }
  }

  public static synchronized void removeProcessLock(final QName processQName) {
    processMutexs.remove(processQName);
  }

  private static synchronized InstanceLock getInstanceLock(final BpelExecution bpelExecution) {
    BpelExecution instance = bpelExecution.getProcessInstance();
    String instanceUuid = instance.getUUID();
    InstanceLock mutex = instanceMutexs.get(instanceUuid);
    return mutex;
  }

  public static InstanceLock lockInstance(final BpelExecution bpelExecution) {
    InstanceLock lock = getInstanceLock(bpelExecution);
    if (lock != null) {
      lock.lock();
    }
    return lock;
  }

  public static void unlockInstance(final InstanceLock instanceLock) {
    if (instanceLock != null) {
      instanceLock.unlock();
    }
  }

  public static synchronized void removeAndUnlockInstance(final String instanceUuid) {
    InstanceLock lock = instanceMutexs.remove(instanceUuid);
    if (lock != null) {
      lock.unlock();
    }
  }

  public static synchronized void createInstanceLock(final String instanceUuid) {
    instanceMutexs.put(instanceUuid, new InstanceLock());
  }

  public static synchronized void createProcessLock(final QName processQName) {
    processMutexs.put(processQName, new ProcessLock());
  }
}