package org.ow2.orchestra.services;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.pvm.activity.Activity;
import org.jbpm.pvm.env.Environment;
import org.jbpm.pvm.env.EnvironmentFactory;
import org.jbpm.pvm.internal.cmd.Command;
import org.jbpm.pvm.internal.cmd.CommandService;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.env.EnvFactoryRepository;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.MutexRepository.InstanceLock;
import org.ow2.orchestra.services.MutexRepository.ProcessLock;
import org.ow2.orchestra.services.impl.ReplierImpl;
import org.ow2.orchestra.services.itf.Publisher;
import org.ow2.orchestra.services.itf.Replier;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.var.Message;

/**
 * base implementation for the services to be published for a deployed bpel
 * process.
 *
 * @author Tom Baeyens, Charles Souillard, Guillaume Porcher
 */
public class Receiver {

  private static class ExecElementToSignal {
    private final BpelExecution bpelExecution;
    private final ReceivingElement receivingElement;

    public BpelExecution getBpelExecution() {
      return this.bpelExecution;
    }
    public ReceivingElement getReceivingElement() {
      return this.receivingElement;
    }
    public ExecElementToSignal(final BpelExecution bpelExecution,
        final ReceivingElement receivingElement) {
      this.bpelExecution = bpelExecution;
      this.receivingElement = receivingElement;
    }
  }

  /** log. */
  private static Logger log = Logger.getLogger(Receiver.class.getName());

  protected long dbid;

  protected int dbversion;

  /**
   * maps ReceiveElementKeys(processQName, portTypeQName, operationName) to a
   * list of executions that are waiting for an incoming message.
   */
  protected Set<BpelExecution> waitingExecutions;

  /**
   * messages stored when there is no execution yet that matches this incoming
   * message.
   */
  protected List<PendingMessage> storedMessages;

  protected OperationKey operationKey;

  protected Receiver() {
    super();
  }

  public Receiver(final OperationKey operationKey) {
    this.operationKey = operationKey;
  }

  protected static class AssociateMessageCommand implements Command<ExecElementToSignal> {
    private final OperationKey operationKey;
    private final Message incomingMessage;
    private final MessageCarrier messageCarrier;

    public AssociateMessageCommand(final OperationKey operationKey,
        final Message incomingMessage, final MessageCarrier messageCarrier) {
      this.operationKey = operationKey;
      this.incomingMessage = incomingMessage;
      this.messageCarrier = messageCarrier;
    }

    public ExecElementToSignal execute(final Environment environment) throws Exception {
      final Publisher servicePublisher = environment.get(Publisher.class);
      final Receiver receiver = servicePublisher.getReceiver(this.operationKey);
      if (receiver == null) {
        throw new OrchestraRuntimeException("impossible to find a receiver corresponding to: " + this.operationKey);
      }
      return receiver.handle(this.incomingMessage, environment, this.messageCarrier);
    }
  }

  protected static class SignalExecCommand implements Command<BpelExecution> {
    private final ExecElementToSignal execToSignal;

    public SignalExecCommand(final ExecElementToSignal execToSignal) {
      this.execToSignal = execToSignal;
    }

    public BpelExecution execute(final Environment environment) throws Exception {
      final Repository repository = environment.get(Repository.class);
      final BpelExecution exec = repository.getUpdatedExecution(this.execToSignal.getBpelExecution());
      if (exec != null) {
        final Map<String, Object> parameters = new HashMap<String, Object>();
        parameters.put("receivingElement",
            repository.getUpdatedReceivingElement(this.execToSignal.getReceivingElement()));
        exec.signal(parameters);
        return exec.getProcessInstance();
      } else {
        return this.execToSignal.getBpelExecution().getProcessInstance();
      }
    }
  }

  protected static class FindExecutionToSignalCommand implements Command<ExecElementToSignal> {
    private final BpelExecution instance;

    public FindExecutionToSignalCommand(final BpelExecution instance) {
      this.instance = instance;
    }

    public ExecElementToSignal execute(final Environment environment) throws Exception {
      final BpelExecution execution = environment.get(Repository.class).getUpdatedExecution(this.instance);
      if (execution != null) {
        return findExecutionToSignal(execution);
      } else {
        return null;
      }
    }
  }
  /**
   * Handle incoming message for the given operation key. Use the given
   * {@link MessageCarrier} to store the reply.
   *
   * @param incomingMessage
   * @param operationKey
   * @param environmentFactory
   * @param messageCarrier
   * @return the execution to which the message was given. <code>null</code>
   *         if there is no execution waiting for this message (the message is
   *         stored).
   */
  public static BpelExecution handle(final Message incomingMessage,
      final OperationKey operationKey, final EnvironmentFactory environmentFactory,
      final MessageCarrier messageCarrier) {
    final CommandService commandService = EnvFactoryRepository.get().get(CommandService.class);

    ExecElementToSignal execToSignal = null;
    // Step 1 : give the message to an existing instance or create a new
    // instance
    // This step is synchronized on the process
    ProcessLock processLock = MutexRepository.lockProcess(operationKey.getProcessQName());
    try {
      execToSignal = commandService.execute(
          new AssociateMessageCommand(operationKey, incomingMessage, messageCarrier)
      );
    } finally {
      MutexRepository.unlockProcess(processLock);
    }
    // Step 2 : continue execution
    // Synchronize on the instance to prevent concurrent modification in
    // separate transactions
    if (execToSignal != null) {
      final BpelExecution currentInstance = execToSignal.getBpelExecution().getProcessInstance();
      InstanceLock lock = MutexRepository.lockInstance(currentInstance);
      try {
        commandService.execute(new SignalExecCommand(execToSignal));
      } finally {
        MutexRepository.unlockInstance(lock);
      }
      ExecElementToSignal execElt = execToSignal;
      // Step 3 : for each child execution, verify that there is no message
      // available.
      // This step is synchronized on the process as we modify the receiver
      // (shared between all instances of a process)
      while (execElt != null && !execElt.getBpelExecution().isEnded()) {
        processLock = MutexRepository.lockProcess(operationKey.getProcessQName());
        try {
          lock = MutexRepository.lockInstance(currentInstance);
          try {
            execElt = commandService.execute(new FindExecutionToSignalCommand(currentInstance));
          } finally {
            MutexRepository.unlockInstance(lock);
          }
        } finally {
          MutexRepository.unlockProcess(processLock);
        }
        if (execElt != null) {
          // Step 4: if a message can be associated with an execution, execute
          // the instance
          // (as for step 2, this is synchronized on the instance)
          lock = MutexRepository.lockInstance(currentInstance);
          try {
            commandService.execute(new SignalExecCommand(execElt));
          } finally {
            MutexRepository.unlockInstance(lock);
          }
        }
      }
      return execToSignal.getBpelExecution();
    }

    return null;
  }

  /**
   * Find if there is a stored message for one of the executions. If such an
   * execution is found, associate the stored message to this execution, and
   * return this execution.
   *
   * @param parent
   *            execution to check (recursive search in the child executions)
   * @return first execution found with a matching message.
   */
  private static ExecElementToSignal findExecutionToSignal(final BpelExecution instance) {
    if (instance == null || instance.isEnded()) {
      return null;
    }
    final Environment environment = Environment.getCurrent();
    ExecElementToSignal execToSignal = null;
    final Set<BpelExecution> waitingExec = new HashSet<BpelExecution>(instance
        .getWaitingExecutions());
    if (waitingExec != null) {
      for (final BpelExecution child : waitingExec) {
        final Activity activity = child.getNode().getBehaviour();
        if (activity instanceof InboundMessageElement) {
          for (final ReceivingElement receivingElement : ((InboundMessageElement)activity).getReceivingElements()) {
            final OperationKey opKey = receivingElement.getOperationKey(child
                .getProcessDefinition().getQName());
            final Receiver receiver = environment.get(Repository.class).getReceiver(
                opKey);
            receiver.addWaitingExecution(child);
            if (execToSignal == null) {
              final Message message = receiver.getMessageFromStoredMessage(child, receivingElement);
              if (message != null) {
                if (!receiver.removeWaitingExecution(child)) {
                  throw new OrchestraRuntimeException("No such waiting execution : " + child);
                }
                associateMessage(child, receivingElement, message);
                execToSignal = new ExecElementToSignal(child, receivingElement);
              }
            }
          }
        }
      }
    }
    return execToSignal;
  }

  /**
   * Add all executions of the given instance that are waiting on a receiving
   * activity to the list waiting activities of the matching receiver.
   *
   * @param parent
   *            the execution
   */
  public static void addWaitingExecutions(final BpelExecution parent) {
    if (parent == null) {
      return;
    }
    final Environment environment = Environment.getCurrent();
    final BpelExecution instance = parent.getProcessInstance();
    final Set<BpelExecution> waitingExec = new HashSet<BpelExecution>(instance
        .getWaitingExecutions());
    if (log.isLoggable(Level.FINE)) {
      log.fine(Receiver.class + ".addWaitingExecutions, execution = " + parent + " waiting=" + waitingExec);
    }
    if (waitingExec != null) {
      for (final BpelExecution child : waitingExec) {
        final Activity activity = child.getNode().getBehaviour();
        if (activity instanceof InboundMessageElement) {
          final InboundMessageElement inboundMessageElement = (InboundMessageElement) activity;
          for (final ReceivingElement receivingElement : inboundMessageElement.getReceivingElements()) {
            final OperationKey opKey = receivingElement.getOperationKey(child
                .getProcessDefinition().getQName());
            final Receiver receiver = environment.get(Repository.class).getReceiver(
                opKey);
            receiver.addWaitingExecution(child);
          }
        }
      }
    }
  }

  /**
   * is called when a message arrives in the engine. Try to find an execution
   * which is waiting for this kind of message (it means operationKeys matches
   * and if the waiting execution activity has correlation check if correlations
   * are OK. If no matching execution was found, try to use the given message to
   * create a new instance of this process if this message can be given to a
   * startActivty being createInstance=true Else, store the message so that it
   * can be used later when an instance will need it. If the given replier is
   * not null, associate it to the corresponding runtimeScope (top level scope
   * if messageExchange is not specified, scopeRuntime declaring the used
   * messageExchange if there is one.
   *
   * @param incomingMessage
   * @param environment
   */

  private ExecElementToSignal handle(final Message incomingMessage,
      final Environment environment, final MessageCarrier messageCarrier) {
    ExecElementToSignal execElementToSignal = null;

    final Repository repository = environment.get(Repository.class);
    final BpelProcess bpelProcess = repository.getProcess(this.operationKey
        .getProcessQName());

    // find a waiting execution using the correlations
    execElementToSignal = getWaitingExecution(incomingMessage, environment);
    // if an execution was waiting for such an incoming message
    if (execElementToSignal != null) {
      if (log.isLoggable(Level.FINE)) {
        log.fine("Matching execution found for incoming msg: " + execElementToSignal.getBpelExecution());
      }
      addReplier(execElementToSignal, environment, messageCarrier);
      return execElementToSignal;
    } else {
      final ReceivingElement startElement = bpelProcess.getStartElement(this.operationKey);
      // if the incoming message matched a start activity
      if (startElement != null) {
        if (log.isLoggable(Level.FINE)) {
          log.fine("found start node " + startElement + " for incoming msg");
        }
        final BpelExecution bpelExecution = (BpelExecution) bpelProcess.beginProcessInstance();
        addWaitingExecutions(bpelExecution);
        final ExecElementToSignal receiveExecution = getWaitingExecution(bpelExecution,
            startElement);
        if (receiveExecution == null) {
          throw new OrchestraRuntimeException(
              "No waiting execution found !, process = "
                  + bpelExecution.getProcessDefinition().getQName());
        }
        associateMessage(receiveExecution.getBpelExecution(), startElement, incomingMessage);
        addReplier(receiveExecution, environment, messageCarrier);
        return receiveExecution;
        // initiaceCS

      } else { // no start node was found
        if (log.isLoggable(Level.FINE)) {
          log.fine("no matching execution found for incoming msg, storing message.");
        }
        storeIncomingMessage(incomingMessage, messageCarrier);
        return null;
      }
    }
  }

  /**
   * Associate the given replier to the corresponding runtimeScope (top level
   * scope if messageExchange is not specified, scopeRuntime declaring the used
   * messageExchange if there is one.
   */
  private void addReplier(final ExecElementToSignal elementToSignal, final Environment environment,
      final MessageCarrier messageCarrier) {
    if (messageCarrier != null) {
      final ReceivingElement receivingElement = elementToSignal.getReceivingElement();
      final BpelExecution bpelExecution = elementToSignal.getBpelExecution();
      final String messageExchange = receivingElement.getMessageExchange();
      final ReplierKey replierKey = new ReplierKey(this.operationKey, messageExchange,
          bpelExecution.getScopeRuntime());
      final Replier replier = createReplier(replierKey);
      MessageCarrierRepository.addMessageCarrier(replierKey, messageCarrier);
      environment.get(Repository.class).storeReplier(replier);
    }
  }

  /**
   * Search for an execution of the given process waiting for a message on the
   * given portType/operation. I one was found and the cs matches (if any),
   * associate this execution to the given message and return this execution.
   *
   * @param incomingMessage
   *            message received we want to give to an execution
   * @param operationKey
   *            operationKey
   * @param environment
   *            environment
   * @return the found execution or null if no one was found
   */
  private ExecElementToSignal getWaitingExecution(final Message incomingMessage,
      final Environment environment) {
    if (this.waitingExecutions == null) {
      if (log.isLoggable(Level.FINE)) {
        log.fine("No waiting executions for Receiver: " + this.getOperationKey());
      }
      return null;
    }
    if (log.isLoggable(Level.FINE)) {
      log.fine("Waiting executions for Receiver " + this.getOperationKey() + ": " + this.waitingExecutions);
    }
    for (final BpelExecution bpelExecution : this.waitingExecutions) {
      // ReceivingElement receivingElement =
      // waitingExecution.getReceivingElement();
      final InboundMessageElement inboundMessageElement = (InboundMessageElement) bpelExecution
          .getNode().getBehaviour();
      for (final ReceivingElement receivingElement : inboundMessageElement.getReceivingElements()) {
        if (log.isLoggable(Level.FINE)) {
          log.fine("Receiving Element: " + receivingElement.getOperationKey(bpelExecution.getProcessDefinition().getQName()) + " / OperationKey=" + this.operationKey);
        }
        if (receivingElement.getOperationKey(bpelExecution.getProcessDefinition().getQName()).equals(this.operationKey)) {
          if (csMatch(incomingMessage, receivingElement, bpelExecution)) {
            if (!this.waitingExecutions.remove(bpelExecution)) {
              throw new OrchestraRuntimeException("Impossible to remove candidate : "
                + bpelExecution);
            }
            associateMessage(bpelExecution, receivingElement, incomingMessage);
            return new ExecElementToSignal(bpelExecution, receivingElement);
          }
        }
      }
    }
    return null;
  }

  /**
   * Search for an execution of the given process instance waiting for a message
   * on the given node. This method is used to associate an incoming message
   * which creates a new instance to the newly created instance.
   *
   * @return the found execution or null if no one was found
   */
  private ExecElementToSignal getWaitingExecution(final BpelExecution parent, final ReceivingElement startElement) {
    if (this.waitingExecutions == null) {
      return null;
    }
    for (final BpelExecution bpelExecution : this.waitingExecutions) {
      if (bpelExecution.getProcessInstance().equals(parent)) {
        final Activity activity = bpelExecution.getNode().getBehaviour();
        if (activity instanceof InboundMessageElement) {
          final InboundMessageElement inboundMessageElement =  (InboundMessageElement) activity;
          if (inboundMessageElement.getReceivingElements().contains(startElement)) {
            if (!this.waitingExecutions.remove(bpelExecution)) {
              throw new OrchestraRuntimeException("Impossible to remove candidate : " + bpelExecution);
            }
            return new ExecElementToSignal(bpelExecution, startElement);
          }
        }
      }
    }
    return null;
  }

  /**
   * Associate the given message to the given execution and startActivity.
   */
  private static void associateMessage(final BpelExecution bpelExecution,
      final ReceivingElement receivingElement, final Message message) {
    bpelExecution.setHasPendingMessage(true);
    bpelExecution.getProcessInstance().getWaitingExecutions().remove(
        bpelExecution);
    receivingElement.receive(message, bpelExecution);

    if (log.isLoggable(Level.FINE)) {
      log.fine(Receiver.class + ".associateMessage, message = " + message);
    }
  }

  /**
   * Store the given message as a message that could be consumed by an activity
   * waiting for a message on the given portType/operation.
   *
   * @param message
   *            message to store
   * @param operationName
   *            operation
   * @param portType
   *            portType
   * @param environment
   *            environment
   *
   */
  private void storeIncomingMessage(final Message message,
      final MessageCarrier messageCarrier) {
    if (this.storedMessages == null) {
      this.storedMessages = new ArrayList<PendingMessage>();
    }
    PendingMessage pendingMessage = null;
    if (messageCarrier != null) {
      MessageCarrierRepository.addPendingMessageCarrier(messageCarrier);
      pendingMessage = new PendingMessage(message, messageCarrier.getUuid());
    } else {
      pendingMessage = new PendingMessage(message, null);
    }
    this.storedMessages.add(pendingMessage);
  }

  // ///////////////////////////////////////////////////////////////////////////////////////

  /**
   * Return true if the given message match the given startActivity for the
   * given execution. It takes in charge cs verification.
   *
   * @param messageToAnalyse
   *            messageToAnalyse
   * @param receivingElement
   *            startActivity
   * @param bpelExecution
   *            bpelExecution
   * @return true if the given message match the given startActivity for the
   *         given execution.
   */
  private boolean csMatch(final Message messageToAnalyse,
      final ReceivingElement receivingElement, final BpelExecution bpelExecution) {
    if (receivingElement.getCorrelations() == null) {
      return true;
    }
    // we calcul the values for each correlationSet
    for (final Correlation correlation : receivingElement.getCorrelations()) {
      if (!correlation.correlationOK(messageToAnalyse, receivingElement,
          bpelExecution)) {
        return false;
      }
    }
    return true;

  }

  /**
   * Search for a stored message matching the given parameters.
   *
   * @param bpelExecution
   *            bpelExecution
   * @param receivingElement
   *            startActivity
   * @return a stored message matching the given parameters. or null.
   */
  private Message getMessageFromStoredMessage(final BpelExecution bpelExecution,
      final ReceivingElement receivingElement) {
    if (this.storedMessages == null) {
      return null;
    }
    for (final PendingMessage pendingMessage : this.storedMessages) {
      final Message message = pendingMessage.getMessage();
      if (csMatch(message, receivingElement, bpelExecution)) {
        final MessageCarrier messageCarrier = MessageCarrierRepository
            .removePendingMessageCarrier(pendingMessage.getMessageCarrierId());
        addReplier(new ExecElementToSignal(bpelExecution, receivingElement), Environment.getCurrent(),
            messageCarrier);
        this.storedMessages.remove(pendingMessage);
        return message;
      }
    }
    return null;
  }

  /**
   * Add a waiting execution in the the list.
   *
   * @param bpelExecution
   *            bpelExecution
   * @param receivingElement
   *            startActivity
   */
  private void addWaitingExecution(final BpelExecution bpelExecution) {
    final ProcessLock processLock = MutexRepository.lockProcess(bpelExecution.getProcessDefinition().getQName());
    try {
      if (this.waitingExecutions == null) {
        this.waitingExecutions = new HashSet<BpelExecution>();
      }
      this.waitingExecutions.add(bpelExecution);
      if (log.isLoggable(Level.FINE)) {
        log.fine(Receiver.class + ".addWaitingExecution, execution = " + bpelExecution);
      }
    } finally {
      MutexRepository.unlockProcess(processLock);
    }
  }

  public boolean removeWaitingExecution(final BpelExecution bpelExecution) {
    final ProcessLock processLock = MutexRepository.lockProcess(bpelExecution.getProcessDefinition().getQName());
    try {
      if (this.waitingExecutions == null) {
        return false;
      }
      return this.waitingExecutions.remove(bpelExecution);
    } finally {
      MutexRepository.unlockProcess(processLock);
    }
  }

  public OperationKey getOperationKey() {
    return this.operationKey;
  }

  private Replier createReplier(final ReplierKey replierKey) {
    return new ReplierImpl(replierKey);
  }

}
